/*
 * main.c
 *
 * Author: Alirio Boaventura
 * alirioboaventura@gmail.com
 *
 */

#include "ProtocolGlobalVariables.h"
#include "SampleData.h"
#include "stdio.h"


int ComposeReaderCommand(Uint8 cmd, Uint8 Q, Uint8 *ReaderCommandBits, Uint8 *NumberOfBits, Uint8 *RN16)
{
	/*	This function composes the bit sequence of a Reader command. CRC should be appended afterward.
	 *	Input:
	 *		cmd - command code
	 *		Q - slot parameter
	 *		ReaderCommandBits - pointer to the Reader command symbol array
	 *		NumberOfBits - number of bits excluding the checksum
	 *		RN16 - pointer to the Transponder RN16 (for the Ack command)
	 *	Output:
	 *		Error code (zero for success, negative for failure)
	 */

	int n = 0;

	if (cmd == CMD_SELECT)
	{
		// Command code: Select, cmd=[1 0 1 0]
		ReaderCommandBits[n++] = 1; // n post-increment
		ReaderCommandBits[n++] = 0;
		ReaderCommandBits[n++] = 1;
		ReaderCommandBits[n++] = 0;

		// Target and Action indicate whether and how a Select modifies a Tags SL or inventoried flag
		// Target: indicates whether the Select modifies a Transponders SL flag or its inventoried flag
		ReaderCommandBits[n++] = 1; // Target: SL
		ReaderCommandBits[n++] = 0;
		ReaderCommandBits[n++] = 0;

		// Action: determines the  behavior of matching and not-matching Transponders
		// on whether to assert or deassert SL or set their inventoried flag to A or B
		ReaderCommandBits[n++] = 0; // Action: assert matching, de-assert non-matching
		ReaderCommandBits[n++] = 0;
		ReaderCommandBits[n++] = 0;

		// MemBank: specifies how a Transponder applies Mask
		ReaderCommandBits[n++] = 0; //EPC memory bank (01)
		ReaderCommandBits[n++] = 1;

		// Pointer: specifies a starting bit address for the Mask comparison
		ReaderCommandBits[n++] = 0; // EPC Start Address
		ReaderCommandBits[n++] = 0;
		ReaderCommandBits[n++] = 1;
		ReaderCommandBits[n++] = 0;
		ReaderCommandBits[n++] = 0;
		ReaderCommandBits[n++] = 0;
		ReaderCommandBits[n++] = 0;
		ReaderCommandBits[n++] = 0;

		// Length: specifies the length of Mask
		ReaderCommandBits[n++] = 0; // 0-length mask causes Transponder to automatically match
		ReaderCommandBits[n++] = 0;
		ReaderCommandBits[n++] = 0;
		ReaderCommandBits[n++] = 0;
		ReaderCommandBits[n++] = 0;
		ReaderCommandBits[n++] = 0;
		ReaderCommandBits[n++] = 0;
		ReaderCommandBits[n++] = 0;

		// Mask: either a FileType (if MemBank=00) or a bit string that a Transponder compares
		// to a memory location that begins at Pointer and ends Length bits later (if MemBank<>00).
		ReaderCommandBits[n++] = 0; // Mask value: 0
		ReaderCommandBits[n++] = 0;
		ReaderCommandBits[n++] = 0;
		ReaderCommandBits[n++] = 0;
		ReaderCommandBits[n++] = 0;
		ReaderCommandBits[n++] = 0;
		ReaderCommandBits[n++] = 0;
		ReaderCommandBits[n++] = 0;

		// Truncate: indicates whether a Transponders reply shall be truncated to those EPC bits that follow Mask
		ReaderCommandBits[n++] = 0; // Disable truncation
	}
	// Query command
	if (cmd == CMD_QUERY)
	{
		// Command code: Query, cmd=[1 0 0 0]
		ReaderCommandBits[n++] = 1; // n post-increment
		ReaderCommandBits[n++] = 0;
		ReaderCommandBits[n++] = 0;
		ReaderCommandBits[n++] = 0;

		// DR (TRcal divide ratio): defines BLF=DR/TRcal
		// Example: DR=8 and TRcal = 50us (in Query Preamble) lead to BLF=160kHz
		ReaderCommandBits[n++] = 0;		// DR=0 -> Divide Ratio = 8

		// M (sub-carrier cycles per symbol)
		ReaderCommandBits[n++] = 0;		// M=[0 0], FM0: 1 cycle per symbol
		ReaderCommandBits[n++] = 0;

		// TRext: defines whether the Transponder prepends a Pilot Tone to its response
		ReaderCommandBits[n++] = 1;		// TRext=1, Transponder responds with a Pilot Tone

		// Sel: chooses which Transponders respond to the Query
		ReaderCommandBits[n++] = 0;		// Sel=[0 0], All Transponders respond
		ReaderCommandBits[n++] = 0;

		// Session: chooses a session for the inventory round (S1, S2, S3 or S4)
		ReaderCommandBits[n++] = 0;		// Session=[0 0], Session S1
		ReaderCommandBits[n++] = 0;

		// Target: selects whether A or B-inventoried Transponders participate in the inventory round
		ReaderCommandBits[n++] = 0;  // Target = A

		// Q: sets the number of available slots N = 2^Q, with Q in the interval [0-15]
		ReaderCommandBits[n++] = (Q & 8)>>3;
		ReaderCommandBits[n++] = (Q & 4)>>2;
		ReaderCommandBits[n++] = (Q & 2)>>1;
		ReaderCommandBits[n++] = (Q & 1);
	}
	if (cmd == CMD_QUERY_REP)
	{
		// Command code: QueryRep, cmd=[0 0]
		ReaderCommandBits[n++] = 0;
		ReaderCommandBits[n++] = 0;
		// Session: S0=[0 0]
		ReaderCommandBits[n++] = 0;
		ReaderCommandBits[n++] = 0;
	}
	if (cmd == CMD_ACK_RN16)
	{
		// Command code: Ack [RN16], cmd=[0 1]
		ReaderCommandBits[0] = 0;
		ReaderCommandBits[1] = 1;
		// Append Transponder RN16
		for(n = 2; n < *NumberOfBits; n++)
		{
			ReaderCommandBits[n] = RN16[n-2];
		}
	}

	*NumberOfBits = n;
	return 0;
}


void AppendCRC5(Uint8 *DataBits, Uint8 *NumberOfBits)
{
	/*	This function computes the 5-bit CRC for a message and appends the checksum to it
	 *	Input:
	 *		DataBits - pointer to the bit array containing the message
	 *		NumberOfBits - number of bits in the message
	 *	Output:
	 *		none
	 */

	// CRC-5 Polynomial: x^5 + x^3 + 1
	Uint16 POLYNOMIAL[5] = {0,1,0,0,1};
	// Uint16 CRC-5 Preset
	Uint16 CRC5_Reg[5] = {0,1,0,0,1};

	Uint8 XORflag = 0;
	int i, j= 0;

	// Compute CRC-5
	while(j < *NumberOfBits)
	{
		// XOR flag
		XORflag = CRC5_Reg[0] ^ *(&DataBits[0] + j);

		// Apply left shift to CRCReg
		for(i = 0; i < 4; i++)
		{
			CRC5_Reg[i] = CRC5_Reg[i+1];
		}
		CRC5_Reg[4] = 0;

		if(XORflag == 1)
		{
			for(i = 0; i < 5; i++){
				CRC5_Reg[i] = CRC5_Reg[i] ^ POLYNOMIAL[i];
			}
		}
		j = j + 1;
	}

	// Append 5-bit CRC to the message
	for(i = 0; i < 5; i++)
	{
		DataBits[*NumberOfBits + i] = CRC5_Reg[i];
	}
	// Update message length
	*NumberOfBits = *NumberOfBits + 5;
}


Uint16 ComputeCRC16(Uint8 *CRC16Poly, Uint8 *CRC16Preset, Uint8 *DataBits, Uint8 NumberOfSymbols)
{
	/*	This function computes the CRC-16 of a message whose address is passed through by DataBits
	 *	Input:
	 *			DataBits - pointer to the data symbol array
	 *			NumberOfSymbols - total number of symbols including the checksum
	 *	Output:
	 *			Error code (zero for success, negative for failure)
	 */

	Uint16 CRC16RegDec = 0;
	Uint8 XORflag = 0, CRC16Reg[16];
	int i, j, n, power = 1;

	// Pre-load CRC register with the preset
	for(i = 0; i < 16 ; i++)
	{
		CRC16Reg[i] = CRC16Preset[i];
	}

	// Compute CRC-16
	for (i = 0; i < NumberOfSymbols; i++) {

		// XOR flag
		XORflag = CRC16Reg[0] ^ DataBits[i];

		// Shift the CRC register left by one bit
		for(j = 0; j < 15; j++)
		{
			CRC16Reg[j] = CRC16Reg[j+1];
		}
		// Insert zero in the LSB
		CRC16Reg[15] = 0;

		//  XOR the CRC register with the polynomial
		if(XORflag == 1)
		{
			for (n = 0; n < 16; n++){
				CRC16Reg[n] ^= CRC16Poly[n];
			}
		}
	}

	// Convert 16-bit CRC array into decimal value
	for (i = 15; i >= 0; i--){
	  CRC16RegDec += CRC16Reg[i]*power;
	  power *= 2;
	}

	// XOR CRC with 0xFFFF (equivalent to ones' complement)
	CRC16RegDec = ~ CRC16RegDec;

	return CRC16RegDec;
}


int CheckCRC16(Uint8 *CRC16Poly, Uint8 *CRC16Preset, Uint8 *DataBits, Uint8 NumberOfBits)
{
	/*	This function verifies the CRC-16 of a message whose address is passed through DataBits
	 *  Success is evaluated based on the final Residue after shifting in the received message plus checksum.
	 *	Input:
	 *			DataBits - pointer to the bit array containing the message
	 *			NumberOfBits - number of bits in the message (checksum included)
	 *	Output:
	 *			Error code (zero for success, negative for failure)
	 */

	Uint16 RcvdCRC = 0, CalcCRC;
	Uint8 CRCBits[16];
	int i, power = 1;

	CalcCRC = ComputeCRC16(CRC16Poly, CRC16Preset, DataBits, NumberOfBits - 16);

	// Get received 16-bit CRC for comparison
	for (i = NumberOfBits - 16; i < NumberOfBits; i++) {
		CRCBits[i -  NumberOfBits + 16] = DataBits[i];
	}
	for (i = 15; i >= 0; i--){
		RcvdCRC += CRCBits[i]*power;
		power *= 2;
	}

	if(RcvdCRC == CalcCRC)
	{
		return OK;
	}
	else
		return CRC16_FAIL;
}


void AppendCRC16(Uint8 *CRC16Poly, Uint8 *CRC16Preset, Uint8 *DataSymbols, Uint8 *NumberOfSymbols)
{
	/*	This function computes the CRC-16 of a message whose addressed
	 *  is passed through DataBits	and appends the checksum to the message
	 *	Input:
	 *		DataBits - pointer to the bit array containing the message
	 *		NumberOfBits - number of bits in the message
	 *	Output:
	 *		none
	 */

	Uint16 CRC16;
	Uint8 CRCBits[16];
	int i;

	// Compute CRC
	CRC16 = ComputeCRC16(CRC16Poly, CRC16Preset, DataSymbols,*NumberOfSymbols);

	// Convert the 16-bit CRC word into array of bits
	for(i = 0; i < 16; i++)
	{
		CRCBits[15-i] = (CRC16 & (0x01 << i)) >> i;
	}

	// Append the 16-bit CRC to the symbol array
	for(i = 0; i < 16; i++)
	{
		DataSymbols[*NumberOfSymbols + i] = CRCBits[i];
	}
	// Update symbol array length
	*NumberOfSymbols = *NumberOfSymbols + 16;
}


Uint16 ComputeCRC16TableMethod(Uint16 crc16, Uint16 *table,
		Uint8 *dataBytes, unsigned int length)
{
	/* This function computes a CRC-16 code based on the table-based byte-wise algorithm
	 * presented in "TI Application Report SLAA221 CRC Implementation With MSP430".
		Input:
				crc16: CRC initial value
				table: pointer to CRC table
				dataBytes: pointer to message
				length: length of message in bytes
		Output:
				16-bit CRC word
	*/
	while(length--)
		crc16 = table[((crc16 >> 8) ^ *dataBytes++)] ^ (crc16 << 8);
	return (crc16 ^ 0xFFFF);
}


void CRC16BuildTable(Uint16 CRC16Poly, Uint16 *ptable)
{
	/* This function generates a 256-word table for the CRC-16 polynomial CRC16Poly
	 * "TI Application Report SLAA221 CRC Implementation With MSP430"
	 *
	 * Input:
	 * 			CRC16Poly: CRC-16 polynomial
	 * 			ptable: pointer to CRC-16 table
	 */

	unsigned int i, j;

	for(i=0; i <= 255; i++)
	{
		ptable[i] = i << 8;
		for(j = 0; j < 8; j++)
			ptable[i] = (ptable[i] << 1) ^ (ptable[i] & 0x8000 ? CRC16Poly : 0);
	}
}


int16 DetectFM0PilotTone1(Uint32 *CPUClockCycles, Uint16 MaxTransitions)
{
	/* This function detects the FM0 pilot tone, by searching for at least 4 consecutive FM0 zeros
	 * Input:
	 * 			CPUClockCyles: array with the durations (in clock cycles) between FM0 transitions
	 * 			MaxTransitions: maximum number of transitions
	 * Output:
	 * 			Initial index of payload data or negative error code if the preamble is not found
	 */

		float CyclesPerSymbol = CPU_CLOCK_FREQ/TARGET_BLF;
		float tol = 0.1*CyclesPerSymbol; // Define a 10% tolerance for the number of CPU cycles per symbol period
		float cycles_half0_max = (CyclesPerSymbol + tol)/2;
		float cycles_half0_min = (CyclesPerSymbol - tol)/2;
		Uint16 k = 0;
		Uint8 S1, S2, S3, S4, S5, S6, S7, S8;

		while(k < MaxTransitions)
		{
			S1 = ((CPUClockCycles[k] > cycles_half0_min) && (CPUClockCycles[k] < cycles_half0_max)); // Half symbol 0
			S2 = ((CPUClockCycles[k + 1] > cycles_half0_min) && (CPUClockCycles[k + 1] < cycles_half0_max));
			S3 = ((CPUClockCycles[k + 2] > cycles_half0_min) && (CPUClockCycles[k + 2] < cycles_half0_max));
			S4 = ((CPUClockCycles[k + 3] > cycles_half0_min) && (CPUClockCycles[k + 3] < cycles_half0_max));
			S5 = ((CPUClockCycles[k + 4] > cycles_half0_min) && (CPUClockCycles[k + 4] < cycles_half0_max));
			S6 = ((CPUClockCycles[k + 5] > cycles_half0_min) && (CPUClockCycles[k + 5] < cycles_half0_max));
			S7 = ((CPUClockCycles[k + 6] > cycles_half0_min) && (CPUClockCycles[k + 6] < cycles_half0_max));
			S8 = ((CPUClockCycles[k + 7] > cycles_half0_min) && (CPUClockCycles[k + 7] < cycles_half0_max));

			if(S1 && S2 && S3 && S4 && S5 && S6 && S7 && S8)
			{
				return k + 8; // return final index of detected FM0 zeros
			}

			k++;
		}	return PILOT_NOT_DETECTED;
}


int16 DetectFM0Preamble1(Uint32 *CPUClockCycles, Uint16 IndexPilotTone, Uint16 MaxTransitions)
{
	/* This function searches for the FM0 preamble pattern and returns the initial index of the payload data
	 * Input:
	 * 			CPUClockCyles: array with the durations (in clock cycles) between FM0 transitions
	 * 			IndexPilotTone: index within pilot tone. If transponder replies with no pilot tone, set IndexPilotTone to zero
	 * 			MaxTransitions: maximum number of transitions allowed
	 * Output:
	 * 			Index of the initial index of the data payload or negative error code if the preamble is not found
	 */

		float CyclesPerSymbol = CPU_CLOCK_FREQ/TARGET_BLF;
		float tol = 0.1*CyclesPerSymbol; // Define a 10% tolerance for the number of CPU cycles per symbol period
		float cycles_v_max = 1.5*(CyclesPerSymbol + tol);  // FM0 violation symbol
		float cycles_v_min = 1.5*(CyclesPerSymbol - tol);
		float cycles_1_max = CyclesPerSymbol + tol;
		float cycles_1_min = CyclesPerSymbol - tol;
		float cycles_half0_max = (CyclesPerSymbol + tol)/2;
		float cycles_half0_min = (CyclesPerSymbol - tol)/2;
		Uint16 k;
		Uint8 S1, S2, S3, S4, S5, S6, S7;

		k = IndexPilotTone;
		while(k < MaxTransitions)
		{
			S1 = ((CPUClockCycles[k] > cycles_1_min) && (CPUClockCycles[k] < cycles_1_max)); // Symbol 1
			S2 = ((CPUClockCycles[k + 1] > cycles_half0_min) && (CPUClockCycles[k + 1] < cycles_half0_max)); // Half symbol 0
			S3 = ((CPUClockCycles[k + 2] > cycles_half0_min) && (CPUClockCycles[k + 2] < cycles_half0_max)); // Half symbol 0
			S4 = ((CPUClockCycles[k + 3] > cycles_1_min) && (CPUClockCycles[k + 3] < cycles_1_max)); // Symbol 1
			S5 = ((CPUClockCycles[k + 4] > cycles_half0_min) && (CPUClockCycles[k + 4] < cycles_half0_max)); // Half symbol 0
			S6 = ((CPUClockCycles[k + 5] > cycles_v_min) && (CPUClockCycles[k + 5] < cycles_v_max)); // Violation symbol V
			S7 = ((CPUClockCycles[k + 6] > cycles_1_min) && (CPUClockCycles[k + 6] < cycles_1_max)); // Symbol 1

			if(S1 && S2 && S3 && S4 && S5 && S6 && S7)
			{
				return k + 7; // return initial index of payload data
			}

			k++;
		}
		return PREAMBLE_NOT_DETECTED;
}


int16 DecodeFM0Data1(Uint32 *CPUClockCycles, Uint16 DataInitIndex,
		Uint16 ExpectNumberBits, Uint8 *DecodedBits)
{
	/* This function implements an FM0 data decoder
	 * Input:
	 * 			CPUClockCyles: array with the durations (in clock cycles) between FM0 transitions
	 * 			DataInitIndex: initial index of payload data in array CPUClockCycles
	 * 			ExpectNumberBits: expected number of bits of the transponder reply
	 * 			DecodedBits: pointer to the array of decoded bits
	 * Output:
	 * 			Error code (zero for success, negative for failure)
	 */

		float CyclesPerSymbol = CPU_CLOCK_FREQ/TARGET_BLF;
		float tol = 0.1*CyclesPerSymbol; // Define a 10% tolerance for the number of CPU cycles per symbol period
		float cycles_1_max = CyclesPerSymbol + tol;
		float cycles_1_min = CyclesPerSymbol - tol;
		float cycles_half0_max = (CyclesPerSymbol + tol)/2;
		float cycles_half0_min = (CyclesPerSymbol - tol)/2;
		Uint16 m = 0, k = DataInitIndex;
		Uint8 S1, S0, PreviousSymbol = SYMB_0, PrevSymbWas1, PrevSymbWasHalf0, PrevSymbWas0;

		while(m < ExpectNumberBits)
		{
			// Half period of Symbol 0 should be in the interval [cycles_half0_min:cycles_half0_max]:
			S0 = ((CPUClockCycles[k] > cycles_half0_min) && (CPUClockCycles[k] < cycles_half0_max));
			// One period of Symbol 1 should be in the interval [cycles_1_min:cycles_1_max]:
			S1 = ((CPUClockCycles[k] > cycles_1_min) && (CPUClockCycles[k] < cycles_1_max));
			PrevSymbWas1 = (PreviousSymbol == SYMB_1);
			PrevSymbWasHalf0 = (PreviousSymbol == HALF_SYMB_0);
			PrevSymbWas0 = (PreviousSymbol == SYMB_0);

			if (S1)
			{
				if (PrevSymbWasHalf0)
				{
					return FM0_CODING_ERROR;
				}
				else if (PrevSymbWas1 || PrevSymbWas0)
				{
					DecodedBits[m++] = 1;
					PreviousSymbol = SYMB_1;
				}
			}
			else if (S0)
			{
				if(PrevSymbWasHalf0)
				{
					DecodedBits[m++] = 0;
					PreviousSymbol = SYMB_0;
				}
				else if (PrevSymbWas1 || PrevSymbWas0)
				{
					PreviousSymbol = HALF_SYMB_0;
				}
			}
			else
			{
				return FM0_CODING_ERROR;
			}

			k++;
		}
	return OK;
}


void Bits2Hex(Uint8 *DataBits, Uint8 NumberOfBits, char *DataChars)
{
	/* This function converts bits to hexadecimal
	 * Input:
	 * 			DataBits: pointer to array of bits
	 * 			NumberOfBits: number of bits
	 * 			DataChars: pointer to array of chars
	 * Output:
	 * 			None
	 */

	Uint16 power, NumberOfChars = NumberOfBits/4;
	char chr;
	int j, i;

	for(i = 0; i < NumberOfChars; i++)
	{
		chr = 0;
		power = 1;
		for (j = 3; j >= 0; j--)
		{
			chr += DataBits[i*4 + j]*power;
			power *= 2;
		}
		DataChars[i] = chr;
	}
}


void Bits2Bytes(Uint8 *DataBits, Uint8 NumberOfBits, Uint8 *DataBytes)
{
	/* This function converts bits to bytes
	 * Input:
	 * 			DataBits: pointer to array of bits
	 * 			NumberOfBits: number of bits
	 * 			DataBytes: pointer to array of bytes
	 * Output:
	 * 			None
	 */
	Uint8 NumberOfBytes, i, j;
	int mask;

	NumberOfBytes = NumberOfBits/8;
	for(i = 0; i <  NumberOfBytes; i++)
	{
		mask = 0x80;
		DataBytes[i] = 0;
		for(j = 0; j < 8; j++)
		{
		  if (DataBits[(i * 8) + j] == 1){
			  DataBytes[i] = DataBytes[i] | mask;
		}
		  mask = mask >> 1;
		}
	 }
}


int main(void) {
	
	int error = 0;
	Uint8 NumberOfBits, NumberOfBytes;
	Uint16 ExpectNumberBits, crc16, CRC16Poly = 0x1021, CRC16Preset = 0xFFFF;
	Uint8 ReaderCommandBits[MAX_COMMAND_LENGTH], DecodedBits[MAX_REPLY_LENGTH], DataBytesEPC1[14];
	Uint8 CRC16Polyb[16] = {0,0,0,1,0,0,0,0,0,0,1,0,0,0,0,1}; 	// CRC-16 polynomial: 1021
	Uint8 CRC16Presetb[16] = {1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1};	// CRC-16 preset: 0xFFFF
	//Uint8 CRC16Presetb[16] = {0,1,0,1,0,0,1,1,0,0,0,0,0,0,1,1};
	Uint16 CRC16Table[256];
	int16 DataInitIndex, IndexPilotTone;
	char DataCharsEPC[32];

	// Generate CRC-16 table. This table can be stored in FLASH memory
	CRC16BuildTable(CRC16Poly, CRC16Table);

	while (1)
	{

		////////// Example 1: Compose Select command and append CRC-16 //////////

		NumberOfBits = 0; // This is updated by the function ComposeReaderCommand()
		error = ComposeReaderCommand(CMD_SELECT, Q_0_SLOTS, ReaderCommandBits, &NumberOfBits, RN16);
		AppendCRC16(CRC16Polyb, CRC16Presetb, ReaderCommandBits, &NumberOfBits);
		// CRC consistency check
		error = CheckCRC16(CRC16Polyb, CRC16Presetb, ReaderCommandBits, NumberOfBits);
		if(error == OK)
		{
			printf("Conformant CRC-16");
		}
		else
			printf("Failed CRC-16");

		////////// Example 2: Compose Query command and append CRC-5 //////////

		NumberOfBits = 0; // This is updated by the function ComposeReaderCommand()
		error = ComposeReaderCommand(CMD_QUERY, Q_0_SLOTS, ReaderCommandBits, &NumberOfBits, RN16);
		AppendCRC5(ReaderCommandBits, &NumberOfBits);

		// TODO: Low-level code for transmitting Reader command and receiving Transponder reply go here

		////////// Example 3: Detect FM0 Pilot Tone and Preamble and decode RN16 data //////////
		IndexPilotTone = DetectFM0PilotTone1(CPUCylcesRN16, MAX_TRANSITIONS);
		// If Transponder uses no Preamble, PreambleIndex should be set to zero
		DataInitIndex = DetectFM0Preamble1(CPUCylcesRN16, IndexPilotTone, MAX_TRANSITIONS);
		ExpectNumberBits = 16;
		if (DataInitIndex >= 0) // check if Preamble was detected successfully
		{
			error = DecodeFM0Data1(CPUCylcesRN16, DataInitIndex, ExpectNumberBits, DecodedBits);
		}

		////////// Example 4: Detect FM0 Pilot Tone and Preamble and decode EPC data //////////
		IndexPilotTone = DetectFM0PilotTone1(CPUCylcesEPC, MAX_TRANSITIONS);
		DataInitIndex = DetectFM0Preamble1(CPUCylcesEPC, IndexPilotTone, MAX_TRANSITIONS);
		ExpectNumberBits = 128;
		if (DataInitIndex >= 0) // check if Preamble was detected successfully
		{
			error = DecodeFM0Data1(CPUCylcesEPC, DataInitIndex, ExpectNumberBits, DecodedBits);
		}
		Bits2Hex(DataBitsEPC1, ExpectNumberBits, DataCharsEPC);

		////////// Example 5: Verify CRC-16 of decoded EPC data //////////
		// Bitwise computation
		error = CheckCRC16(CRC16Polyb, CRC16Presetb, DataBitsEPC1, ExpectNumberBits);
		if(error == OK)
		{
			printf("Conformant CRC-16");
		}
		else
			printf("Failed CRC-16");

		// Byte-wise computation
		NumberOfBits = ExpectNumberBits - 16;
		NumberOfBytes = NumberOfBits/8;
		Bits2Bytes(DataBitsEPC1, ExpectNumberBits, DataBytesEPC1);
		// Compute CRC table
		CRC16BuildTable(CRC16Poly, CRC16Table);
		// Compute CRC-16 checksum
		crc16 = ComputeCRC16TableMethod(CRC16Preset, CRC16Table, DataBytesEPC1, NumberOfBytes);
		if(crc16  == 0x615D)
		{
			printf("Conformant CRC-16");
		}
		else
			printf("Failed CRC-16");

	}

	return 0;
}
